/*
  +----------------------------------------------------------------------+
  | Portable System Library                                              |
  | Author: Sukchan Lee (acetcom@postech.ac.kr)                          |
  +----------------------------------------------------------------------+

  $Id: string.c 113 2004-04-06 06:02:40Z acetcom $
*/

#include "psys/common.h"
#include "psys/memory.h"
#include "psys/string.h"
#include "psys/errno.h"

#if !HAVE_MEMCHR
void *memchr(const void *s, int c, size_t n)
{
    const char *cp;

    for (cp = s; n > 0; n--, cp++) 
    {
        if (*cp == c)
        {
            return (char *)cp;
        }
    }

    return NULL;
}
#endif

#if !HAVE_STRSTR
char *strstr(char *s1, char *s2)
{
    char *p1, *p2;

    if (*s2 == '\0') 
    {
        return s1;
    }

    while ((s1 = strchr(s1, *s2)) != NULL) 
    {
        p1 = s1;
        p2 = s2;

        while (*++p1 == *++p2) 
        {
            if (*p1 == '\0') 
            {
                return s1;
            }
        }

        if (*p2 == '\0') 
        {
            break;
        }

        s1++;
    }

    return s1;
}
#endif

#if !HAVE_STRLCPY
size_t strlcpy(char *dst, const char *src, size_t n)
{
    char *d = dst;
    const char *s = src;
    size_t i = n;

    if (dst == NULL || src == NULL)
    {
        return 0;
    }

    if (i != 0 && --i != 0)
    {
        do
        {
            char c = *s++;

            *d++ = c;
            if (c == 0)
            {
                break;
            }
        }
        while (--i != 0);
    }

    if (i == 0)
    {
        if (n != 0)
        {
            *d = 0;
        }
        while (*s++);
    }

    return s - src - 1;
}
#endif

#if !HAVE_STRLCAT
size_t strlcat(char *dst, const char *src, size_t n)
{
    char *d = dst;
    const char *s = src;
    size_t bytes_left = n;
    size_t dlength;

    if (dst == NULL || src == NULL)
    {
        return 0;
    }

    while (*d != 0 && bytes_left-- != 0)
    {
        d++;
    }

    dlength = d - dst;
    bytes_left = n - dlength;

    if (bytes_left == 0)
    {
        return dlength + strlen(s);
    }

    while (*s != 0)
    {
        if (bytes_left != 1)
        {
            *d++ = *s;
            bytes_left--;
        }
        s++;
    }

    *d = 0;

    return dlength + (s - src);
}
#endif

#if (!HAVE_STRCASECMP && !HAVE_STRICMP)
int strcasecmp(const char *s1, const char *s2)
{
    const char *p = s1;
    const char *q = s2;

    for (p = s1, q = s2; *p && *q; p++, q++) 
    {
        int diff = tolower(*p) - tolower(*q);

        if (diff)
        {
            return diff;
        }
    }

    if (*p)
    {
        return 1;
    }
    if (*q)
    {
        return -1;
    }

    return 0;
}
#endif

#if (!HAVE_STRNCASECMP && !HAVE_STRNICMP)
int strncasecmp(const char *s1, const char *s2, size_t n)
{
    const char *p = s1;
    const char *q = s2;

    for (p = s1, q = s2;; p++, q++) 
    {
        int diff;
        
        if (p == s1 + n)
        {
            return 0;
        }
        if (!(*p && *q))
        {
            return *p - *q;
        }

        diff = tolower(*p) - tolower(*q);

        if (diff)
        {
            return diff;
        }
    }
}
#endif

#if !HAVE_VSNPRINTF
int vsnprintf(char *str, size_t str_m, const char *fmt, va_list ap)
{
    size_t str_l = 0;
    const char *p = fmt;

    if (!p) 
    {
        p = "";
    }
    while (*p) 
    {
        if (*p != '%') 
        {
            const char *q = strchr(p + 1, '%');
            size_t n = !q ? strlen(p) : (q - p);
            if (str_l < str_m) 
            {
                size_t avail = str_m - str_l;
                memcpy(str + str_l, p, (n > avail ? avail : n));
            }
            p += n; str_l += n;
        } 
        else 
        {
            const char *starting_p;
            size_t min_field_width = 0, precision = 0;
            int zero_padding = 0, precision_specified = 0, justify_left = 0;
            int alternate_form = 0, force_sign = 0;
            int space_for_positive = 1; 
            char length_modifier = '\0';
            char tmp[32];
            const char *str_arg;
            size_t str_arg_l;
            unsigned char uchar_arg;
            size_t number_of_zeros_to_pad = 0;
            size_t zero_padding_insertion_ind = 0;
            char fmt_spec = '\0';

            str_arg = NULL;
            starting_p = p; p++;  

            while (*p == '0' || *p == '-' || *p == '+' || 
                    *p == ' ' || *p == '#' || *p == '\'') 
            {
                switch (*p) 
                {
                    case '0': 
                        zero_padding = 1; 
                        break;
                    case '-': 
                        justify_left = 1; 
                        break;
                    case '+': 
                        force_sign = 1; 
                        space_for_positive = 0; 
                        break;
                    case ' ': 
                        force_sign = 1;
                        break;
                    case '#': 
                        alternate_form = 1; 
                        break;
                    case '\'': 
                        break;
                }
                p++;
            }

            if (*p == '*') 
            {
                int j;

                p++; j = va_arg(ap, int);
                if (j >= 0) 
                {
                    min_field_width = j;
                }
                else 
                { 
                    min_field_width = -j; justify_left = 1; 
                }
            } 
            else if (isdigit((int)(*p))) 
            {
                unsigned int uj = *p++ - '0';

                while (isdigit((int)(*p))) 
                {
                    uj = 10 * uj + (unsigned int)(*p++ - '0');
                }
                min_field_width = uj;
            }

            if (*p == '.') 
            {
                p++; precision_specified = 1;
                if (*p == '*') 
                {
                    int j = va_arg(ap, int);

                    p++;
                    if (j >= 0) 
                    {
                        precision = j;
                    }
                    else 
                    {
                        precision_specified = 0; precision = 0;
                    }
                } 
                else if (isdigit((int)(*p))) 
                {
                    unsigned int uj = *p++ - '0';
                    while (isdigit((int)(*p))) 
                    {
                        uj = 10 * uj + (unsigned int)(*p++ - '0');
                    }
                    precision = uj;
                }
            }

            if (*p == 'h' || *p == 'l') 
            {
                length_modifier = *p; p++;
                if (length_modifier == 'l' && *p == 'l') 
                {
                    length_modifier = '2';
                    p++;
                }
            }

            fmt_spec = *p;

            switch (fmt_spec) 
            {
                case 'i': 
                    fmt_spec = 'd'; 
                    break;
                case 'D': 
                    fmt_spec = 'd'; 
                    length_modifier = 'l'; 
                    break;
                case 'U': 
                    fmt_spec = 'u'; 
                    length_modifier = 'l'; 
                    break;
                case 'O': 
                    fmt_spec = 'o'; 
                    length_modifier = 'l'; 
                    break;
                default: 
                    break;
            }

            switch (fmt_spec) 
            {
                case '%': 
                case 'c': 
                case 's':
                    length_modifier = '\0';          
                    str_arg_l = 1;
                    switch (fmt_spec) 
                    {
                        case '%': 
                            str_arg = p; 
                            break;
                        case 'c': 
                        {
                            int j = va_arg(ap, int);
                            uchar_arg = (unsigned char)j;   
                            str_arg = (const char *)&uchar_arg;
                            break;
                        }
                        case 's':
                            str_arg = va_arg(ap, const char *);
                            if (!str_arg) 
                            {
                                str_arg_l = 0;
                            }
                            else if (!precision_specified) 
                            {
                                str_arg_l = strlen(str_arg);
                            }
                            else if (precision == 0) 
                            {
                                str_arg_l = 0;
                            }
                            else 
                            {
                                const char *q = memchr(str_arg, '\0',
                                precision <= 0x7fffffff ? 
                                    precision : 0x7fffffff);
                                str_arg_l = !q ? precision : (q - str_arg);
                            }
                            break;
                        default: 
                            break;

                    }
                    break;

                case 'd': 
                case 'u': 
                case 'o': 
                case 'x': 
                case 'X': 
                case 'p': 
                {
                    int arg_sign = 0;
                    int int_arg = 0;  unsigned int uint_arg = 0;
                    long int long_arg = 0;  unsigned long int ulong_arg = 0;
                    void *ptr_arg = NULL;
                    long long int long_long_arg = 0;
                    unsigned long long int ulong_long_arg = 0;

                    if (fmt_spec == 'p') 
                    {
                        length_modifier = '\0';
                        ptr_arg = va_arg(ap, void *);
                        if (ptr_arg != NULL) 
                        {
                            arg_sign = 1;
                        } 
                    }
                    else if (fmt_spec == 'd') 
                    {
                        switch (length_modifier) 
                        {
                            case '\0':
                            case 'h':
                                int_arg = va_arg(ap, int);
                                if (int_arg > 0) 
                                {
                                    arg_sign =  1;
                                }
                                else if (int_arg < 0) 
                                {
                                    arg_sign = -1;
                                }
                                break;
                            case 'l':
                                long_arg = va_arg(ap, long int);
                                if (long_arg > 0) 
                                {
                                    arg_sign =  1;
                                }
                                else if (long_arg < 0)
                                {
                                    arg_sign = -1;
                                }
                                break;
                            case '2':
                                long_long_arg = va_arg(ap, long long int);
                                if (long_long_arg > 0) 
                                {
                                    arg_sign =  1;
                                }
                                else if (long_long_arg < 0) 
                                {
                                    arg_sign = -1;
                                }
                                break;
                        }
                    } 
                    else 
                    {
                        switch (length_modifier) 
                        {
                            case '\0':
                            case 'h':
                                uint_arg = va_arg(ap, unsigned int);
                                if (uint_arg) 
                                {
                                    arg_sign = 1;
                                }
                                break;
                            case 'l':
                                ulong_arg = va_arg(ap, unsigned long int);
                                if (ulong_arg) 
                                {
                                    arg_sign = 1;
                                }
                                break;
                            case '2':
                                ulong_long_arg = 
                                    va_arg(ap, unsigned long long int);
                                if (ulong_long_arg) 
                                {
                                    arg_sign = 1;
                                }
                                break;
                        }
                    }

                    str_arg = tmp; str_arg_l = 0;
                    if (precision_specified) 
                    {
                        zero_padding = 0;
                    }

                    if (fmt_spec == 'd') 
                    {
                        if (force_sign && arg_sign >= 0)
                        {
                            tmp[str_arg_l++] = 
                                space_for_positive ? ' ' : '+';
                        }
                    } 
                    else if (alternate_form) 
                    {
                        if (arg_sign != 0 && (fmt_spec == 'x' || 
                                    fmt_spec == 'X'))
                        { 
                            tmp[str_arg_l++] = '0'; 
                            tmp[str_arg_l++] = fmt_spec; 
                        }
                    }

                    zero_padding_insertion_ind = str_arg_l;
                    if (!precision_specified) 
                    {
                        precision = 1;
                    }
                    if (precision == 0 && arg_sign == 0) 
                    {
                    } 
                    else 
                    {
                        char f[5]; int f_l = 0;
                        f[f_l++] = '%';

                        if (!length_modifier) 
                        { 
                        }
                        else if (length_modifier=='2') 
                        { 
                            f[f_l++] = 'l'; f[f_l++] = 'l'; 
                        }
                        else 
                        {
                            f[f_l++] = length_modifier;
                        }
                        f[f_l++] = fmt_spec; f[f_l++] = '\0';
                        if (fmt_spec == 'p') 
                        {
                            str_arg_l += sprintf(tmp+str_arg_l, f, ptr_arg);
                        }
                        else if (fmt_spec == 'd') 
                        { 
                            switch (length_modifier) 
                            {
                                case '\0':
                                case 'h': 
                                    str_arg_l += sprintf(
                                        tmp + str_arg_l, f, int_arg);  
                                    break;
                                case 'l': 
                                    str_arg_l += sprintf(
                                        tmp + str_arg_l, f, long_arg); 
                                    break;
                                case '2': 
                                    str_arg_l += sprintf(
                                        tmp + str_arg_l,f,long_long_arg); 
                                    break;
                            }
                        } 
                        else 
                        {
                            switch (length_modifier) 
                            {
                                case '\0':
                                case 'h': 
                                    str_arg_l += sprintf(
                                        tmp + str_arg_l, f, uint_arg);  
                                    break;
                                case 'l': 
                                    str_arg_l += sprintf(
                                        tmp + str_arg_l, f, ulong_arg); 
                                    break;
                                case '2': 
                                    str_arg_l += sprintf(
                                        tmp + str_arg_l,f,ulong_long_arg);
                                    break;
                            }
                        }

                        if (zero_padding_insertion_ind < str_arg_l &&
                            tmp[zero_padding_insertion_ind] == '-') 
                        {
                            zero_padding_insertion_ind++;
                        }
                        if (zero_padding_insertion_ind+1 < str_arg_l &&
                            tmp[zero_padding_insertion_ind] == '0' &&
                            (tmp[zero_padding_insertion_ind+1] == 'x' ||
                            tmp[zero_padding_insertion_ind+1] == 'X') ) 
                        {
                            zero_padding_insertion_ind += 2;
                        }
                    }
                    { 
                        size_t num_of_digits = 
                            str_arg_l - zero_padding_insertion_ind;
                        if (alternate_form && fmt_spec == 'o' && 
                            !(zero_padding_insertion_ind < str_arg_l && 
                            tmp[zero_padding_insertion_ind] == '0')) 
                        {
                            if (!precision_specified || 
                                    precision < num_of_digits + 1) 
                            {
                                precision = num_of_digits + 1; 
                                precision_specified = 1;
                            }
                        }
                        if (num_of_digits < precision) 
                        {
                            number_of_zeros_to_pad = 
                                precision - num_of_digits;
                        }
                    }

                    if (!justify_left && zero_padding) 
                    {
                        int n = min_field_width - 
                            (str_arg_l + number_of_zeros_to_pad);
                        if (n > 0) 
                        {
                            number_of_zeros_to_pad += n;
                        }
                    }
                    break;
                }
                default: 
                    zero_padding = 0;
                    justify_left = 1; min_field_width = 0;
                    str_arg = p; str_arg_l = 0;
                    if (*p) 
                    {
                        str_arg_l++;  
                    }
                    break;
            }

            if (*p) 
            {
                p++;
            }
            if (!justify_left) 
            {
                int n = min_field_width - (str_arg_l + number_of_zeros_to_pad);
                if (n > 0) 
                {
                    if (str_l < str_m) 
                    {
                        size_t avail = str_m-str_l;
                        memset(str + str_l, (zero_padding ? '0' : ' ' ), 
                                (n > avail ? avail : n));
                    }
                    str_l += n;
                }
            }

            if (number_of_zeros_to_pad <= 0) 
            {
                zero_padding_insertion_ind = 0;
            } 
            else 
            {
                int n = zero_padding_insertion_ind;
                if (n > 0) 
                {
                    if (str_l < str_m) 
                    {
                        size_t avail = str_m - str_l;
                        memcpy(str + str_l, str_arg, (n > avail ? avail : n));
                    }
                    str_l += n;
                }
                n = number_of_zeros_to_pad;
                if (n > 0) 
                {
                    if (str_l < str_m) 
                    {
                        size_t avail = str_m - str_l;
                        memset(str + str_l, '0', (n > avail ? avail : n));
                    }
                    str_l += n;
                }
            }

            { 
                int n = str_arg_l - zero_padding_insertion_ind;
                if (n > 0) 
                {
                    if (str_l < str_m) 
                    {
                        size_t avail = str_m - str_l;
                        memcpy(str + str_l, 
                                str_arg + zero_padding_insertion_ind,
                                (n > avail ? avail :n ));
                    }
                    str_l += n;
                }
            }

            if (justify_left)
            {
                int n = min_field_width - (str_arg_l+number_of_zeros_to_pad);
                if (n > 0) 
                {
                    if (str_l < str_m) 
                    {
                        size_t avail = str_m - str_l;
                        memset(str + str_l, ' ', (n > avail ? avail : n));
                    }
                    str_l += n;
                }
            }
        }
    }

    if (str_m > 0) 
    {
        str[str_l <= str_m - 1 ? str_l : str_m - 1] = '\0';
    }

    return (int)str_l;
}
#endif

#if !HAVE_SNPRINTF
int snprintf(char *str, size_t str_m, const char *fmt, ...)
{
	va_list ap;
	int str_l;

	va_start(ap, fmt);
	str_l = vsnprintf(str, str_m, fmt, ap);
	va_end(ap);

	return str_l;
}
#endif

char *psys_strdup(psys_mem_pool_t *p, const char *s)
{
    char *rv;
    size_t len;

    if (s == NULL) 
    {
        return NULL;
    }

    len = strlen(s) + 1;
    rv = psys_malloc(p, len);
    memcpy(rv, s, len);

    return rv;
}

char *psys_strndup(psys_mem_pool_t *p, const char *s, size_t n)
{
    char *rv;
    const char *end;

    if (s == NULL) 
    {
        return NULL;
    }

    end = memchr(s, '\0', n);

    if (end != NULL)
    {
        n = end - s;
    }

    rv = psys_malloc(p, n + 1);
    memcpy(rv, s, n);
    rv[n] = '\0';

    return rv;
}

void *psys_memdup(psys_mem_pool_t *p, const void *m, size_t n)
{
    void *rv;

    if (m == NULL)
    {
        return NULL;
    }

    rv = psys_malloc(p, n);
    memcpy(rv, m, n);

    return rv;
}

char *psys_itoa(psys_mem_pool_t *p, int n)
{
    const int buf_size = sizeof(int) * 3 + 2;
    char *buf = psys_malloc(p, buf_size);
    char *start = buf + buf_size - 1;
    int negative;

    if (n < 0) 
    {
        negative = 1;
        n = -n;
    }
    else 
    {
        negative = 0;
    }

    *start = 0;

    do 
    {
        *--start = '0' + (n % 10);
        n /= 10;
    } 
    while (n);

    if (negative) 
    {
        *--start = '-';
    }

    return start;
}

char *psys_ltoa(psys_mem_pool_t *p, long n)
{
    const int buf_size = sizeof(long) * 3 + 2;
    char *buf = psys_malloc(p, buf_size);
    char *start = buf + buf_size - 1;
    int negative;

    if (n < 0) 
    {
        negative = 1;
        n = -n;
    }
    else 
    {
        negative = 0;
    }

    *start = 0;

    do 
    {
        *--start = (char)('0' + (n % 10));
        n /= 10;
    } 
    while (n);

    if (negative) 
    {
        *--start = '-';
    }

    return start;
}

p_int64_t psys_strtoi64(const char *buf, char **end, int base)
{
#if PSYS_HAVE_STRTOI64_FUNCTION
    return PSYS_STRTOI64_FUNCTION(buf, end, base);
#else
    const char *s;
    p_int64_t acc;
    p_int64_t val;
    int neg, any;
    char c;

    s = buf;
    do 
    {
        c = *s++;
    } while (isspace(c));
    if (c == '-') 
    {
        neg = 1;
        c = *s++;
    } 
    else 
    {
        neg = 0;
        if (c == '+')
        {
            c = *s++;
        }
    }
    if ((base == 0 || base == 16) && c == '0' && (*s == 'x' || *s == 'X')) 
    {
	    c = s[1];
	    s += 2;
	    base = 16;
    }
    if (base == 0)
    {
        base = c == '0' ? 8 : 10;
    }
    acc = any = 0;
    if (base < 2 || base > 36) 
    {
        errno = EINVAL;
        if (end != NULL)
        {
            *end = (char *)(any ? s - 1 : buf);
        }
        return acc;
    }

    val = 0;
    for (;; c = *s++) 
    {
        if (c >= '0' && c <= '9')
        {
            c -= '0';
        }
        else if (c >= 'A' && c <= 'Z')
        {
            c -= 'A' - 10;
        }
        else if (c >= 'a' && c <= 'z')
        {
            c -= 'a' - 10;
        }
        else
        {
            break;
        }
        if (c >= base)
        {
            break;
        }
        val *= base;
        if ((any < 0) || (neg && (val > acc || (val -= c) > acc))
           || (!neg && (val < acc || (val += c) < acc)))
        {
            any = -1;
        } 
        else 
        {
            acc = val;
            any = 1;
        }
    }

    if (any < 0) 
    {
        acc = neg ? PSYS_INT64_MIN : PSYS_INT64_MAX;
        errno = ERANGE;
    } 
    else if (!any) 
    {
        errno = EINVAL;
    }
    if (end != NULL)
    {
        *end = (char *)(any ? s - 1 : buf);
    }

    return acc;
#endif
}

p_int64_t psys_atoi64(const char *buf)
{
    return psys_strtoi64(buf, NULL, 0);
}
